#!!

Create full swing bar from swing
================================

Input:
Swing - list<pos, value>
Beats - list<pos, barnum, beatnum>

Output:
Swing - list<pos, value> ;; All values are filled out.


Create swing filledout list
===========================

Input:
======

PrevSwing - list<pos, value> ;; Swing of previous track. If a value is 0, it means that automatic repeat of swing stops. (Same thing also happens if the measure changes)
                             ;; Perhaps BlockSwing is better than PrevTrackSwing? (no, because then we would need a custom "BlockSwing" track). 
                             ;; But perhaps FirstTrackSwing instead of PrevTrackSwing?
                             ;;    + If adding swing at Track #2, there won't be any swing on Track #3, unless there is swing on Track #1
                             ;;    - Track #1 will have a special meaning, but maybe that is not such a bad thing.
                             ;;    - Prev track could give some interesting results.
                             ;;    + If adding a custom BlockSwing track later, the behavior will be more similar and it might be easier to load older songs.
                             ;;    - If using previous track, it doesn't make very much sense to make a custom BlockSwing track.
                             ;;    - It might be more common that you want to use previous track.
                             ;;    + It might be more common to use a default swing track.
                             ;;    + Probably less confusing, but not sure. For sure less confusing to use custom BlockSwing track though.
                             ;;    + A BlockSwing track is more excplicit (not sure)
                             ;;    - The missing numbers are filled in automatically anyway, so it doesn't matter that much if it's less confusing.
                             ;;    + If changing swing of track 3, it's likely that you don't want to change swing of track 4.
Swing     - list<pos, value> ;; Same behavior if value is 0 as above.
Beats     - list<pos, barnum, beatnum>
LPBs      - list<pos, lpb>
DefaultLPB - integer
NumLines  - integer

Output:
list<pos, value> ;; A value can not be 0, and all automatic repeat of swing has been filled out.


!!#

::::::::::::::::::::::::::::::::::::::
;;;;;;;;;;;; BAR :::::::::::::::::::::
::::::::::::::::::::::::::::::::::::::

(define (create-beat line num)
  (list line num))

(define (get-beat-line beat)
  (car beat))

(define (get-beat-num beat)
  (cadr beat))

(define-struct bar
  :barnum
  :beats)

(define (get-first-line-in-bar bar)
  (get-beat-line 
   (car (bar :beats))))

(define (get-last-line-in-bar bar)
  (get-beat-line 
   (car (bar :beats))))




::::::::::::::::::::::::::::::::::::::
;;;;;;;;;;;; SWING :::::::::::::::::::
::::::::::::::::::::::::::::::::::::::

(define (create-swing line weight)
  (list line weight))

(define (get-swing-line swing)
  (car swing))

(define (get-swing-weight swing)
  (cadr swing))


(define-struct bar-swing
  :barnum
  :swings
  :num-lines)


;;(define (bars-have-same-signatures? prev-bar prev-bar-length bar curr-bar-length)
;;  (= prev-bar-length curr-bar-length)) ;; Probably good enough.


(define *test-shuffle-swing* '((0 1)
                               (2 2)
                               (4 1)
                               (6 2)
                               (8 1)
                               (10 2)
                               (12 1)
                               (14 2)))

(define *test-4/4-beats* (list (make-bar :barnum 0 :beats '(( 0 0)
                                                            ( 4 1)
                                                            ( 8 2)
                                                            (12 3)))
                               (make-bar :barnum 1 :beats '((16 0)
                                                            (20 1)
                                                            (24 2)
                                                            (28 3)))))


#||
;; Currently just moves the first swing in the bar to the start of the bar. (not needed, incorportated into 'group-swing-by-bar' instead)
(define (add-bar-start-to-swing bars swing)
  (if (null? bars)
      '()
      (let* ((bar (car bars))
             (next-bar (cl-cadr bars))
             (first-line (car bar))
             (next-line (or (and next-bar (car (cadr next-bar)))
                            num-lines)))

        )))

(***assert*** (add-bar-start-swings '((0 (0 0))
                                      (1 (2 0)))
                                    '((0 1)
                                      (1 2)
                                      (3 5)))
              '((0 1)
                (1 2)
                (2 5)))
(***assert*** (add-bar-start-swings '((0 (0 0))
                                      (1 (2 0)))
                                    '((0 1)
                                      (1 2)
                                      (3 5)
                                      (7/2 6)))
              '((0 1)
                (1 2)
                (2 5)
                (7/2 6)))
||#           

(define (group-swing-by-bar bars swings num-lines)
  (if (null? bars)
      '()
      (let* ((bar (car bars))
             (next-bar (begin (c-display ":bar" bar ":bars" bars) (cl-cadr bars)))
             (barnum (bar :barnum))
             (first-line (get-first-line-in-bar bar))
             (next-line (if next-bar
                            (get-first-line-in-bar next-bar)
                            num-lines)))
        (let loop ((swings swings)
                   (result '())
                   (is-first-bar #t))
          (define (gotit)
            ;;(c-display "gotit" first-line result)
            (if (null? result)
                (group-swing-by-bar (cdr bars) swings num-lines)
                (cons (make-bar-swing :barnum barnum
                                      :swings (map (lambda (swing)
                                                     (create-swing (- (get-swing-line swing)
                                                                      first-line)
                                                                   (get-swing-weight swing)))
                                                   (reverse result))
                                      :num-lines (- next-line first-line))
                      (group-swing-by-bar (cdr bars) swings num-lines))))
          ;;(c-display "swing" swing)
          (c-display ":bar" bar (cl-cadr next-bar) ":first-line" first-line ":swings" swings ":result" result ":next-line" next-line)
          (if (null? swings)
              (gotit)
              (let* ((swing (car swings))
                     (swing-line (get-swing-line swing)))
                (if (>= swing-line next-line)
                    (gotit)
                    (loop (cdr swings)
                          (cons (if (and is-first-bar
                                         (> swing-line first-line))
                                    (create-swing first-line
                                                  (get-swing-weight swing))
                                    swing)
                                result)
                          #f))))))))
        
(pretty-print (group-swing-by-bar (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                        (make-bar :barnum 1 :beats (list (create-beat 2 0))))
                                  (list (create-swing 0 1)
                                        (create-swing 1 2)
                                        (create-swing 3 5))
                                  4))
        

(***assert*** (group-swing-by-bar (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                        (make-bar :barnum 1 :beats (list (create-beat 2 0))))
                                  (list (create-swing 0 1)
                                        (create-swing 1 2)
                                        (create-swing 2 4)
                                        (create-swing 3 5))
                                  4)
              (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                            (create-swing 1 2))
                                    :num-lines 2)
                    (make-bar-swing :barnum 1 :swings (list (create-swing 0 4)
                                                            (create-swing 1 5))
                                    :num-lines 2)))


(***assert*** (group-swing-by-bar (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                        (make-bar :barnum 1 :beats (list (create-beat 2 0))))
                                  (list (create-swing 0 1)
                                        (create-swing 1 2)
                                        (create-swing 3 5))
                                  4)
              (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                            (create-swing 1 2))
                                    :num-lines 2)
                    (make-bar-swing :barnum 1 :swings (list (create-swing 0 5))
                                    :num-lines 2)))


#||
(***assert*** (group-swing-by-bar *test-4/4-beats*
                                  *test-shuffle-swing*
                                  64)
              `((0 ,*test-shuffle-swing*)))
||#


;; Add autoinserted swings.
(define (create-filledout-swings bars global-swings track-swings num-lines)  
  (let loop ((bars bars)
             (prev-bar #f)
             (curr-swing #f)
             (global-swings global-swings) ;;(group-swing-by-bar bars global-swing num-lines))
             (track-swings track-swings)) ;;(group-swing-by-bar bars track-swing num-lines)))
    (cond ((null? bars)
           (assert (null? global-swings))
           (assert (null? track-swings))
           '())

          ((and (not curr-swing)
                (null? global-swings)
                (null? track-swings))
           '())

          (else

           (define bar (car bars))
           (define barnum (bar :barnum))

           (while (and (not (null? global-swings))
                       (< ((car global-swings) :barnum) barnum))
             (set! global-swings (cdr global-swings)))

           (define global-swing (cl-car global-swings))
           (define track-swing (cl-car track-swings))

           (cond ((and track-swing
                       (= barnum (track-swing :barnum)))
                  (cons track-swing
                        (loop (cdr bars)
                              bar
                              track-swing
                              global-swings
                              (cdr track-swings))))

                 ((and global-swing
                       (= barnum (global-swing :barnum)))
                  (cons global-swing
                        (loop (cdr bars)
                              bar
                              global-swing
                              (cdr global-swings)
                              track-swings)))

                 ((and curr-swing
                       (let* ((curr-line (get-first-line-in-bar bar))
                              (next-bar (cl-cadr bars))
                              (next-line (if next-bar
                                             (get-first-line-in-bar next-bar)
                                             num-lines))
                              (curr-bar-length (- next-line
                                                  curr-line)))
                         (= (curr-swing :num-lines)
                            curr-bar-length)))

                  (cons (copy-bar-swing curr-swing :barnum barnum)
                        (loop (cdr bars)
                              bar
                              curr-swing
                              global-swings
                              track-swings)))

                 (else
                  (loop (cdr bars)
                        bar
                        #f
                        global-swings
                        track-swings)))))))



(***assert*** (create-filledout-swings (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                             (make-bar :barnum 1 :beats (list (create-beat 2 0)))
                                             (make-bar :barnum 2 :beats (list (create-beat 4 0))))
                                       (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                                                     (create-swing 1 2))
                                                             :num-lines 2))
                                       (list (make-bar-swing :barnum 2 :swings (list (create-swing 0 2)
                                                                                     (create-swing 1 1))
                                                             :num-lines 2))
                                       6)
              (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                            (create-swing 1 2))
                                    :num-lines 2)
                    (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                            (create-swing 1 2))
                                    :num-lines 2)
                    (make-bar-swing :barnum 2 :swings (list (create-swing 0 2)
                                                            (create-swing 1 1))
                                    :num-lines 2)))

;; Test that we don't repeat swing when there is a signature change. (actually, when the bar changes number of lines)
(***assert*** (create-filledout-swings (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                             (make-bar :barnum 1 :beats (list (create-beat 2 0)))
                                             (make-bar :barnum 2 :beats (list (create-beat 4 0))))
                                       (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                                                     (create-swing 1 2))
                                                             :num-lines 2))
                                       (list (make-bar-swing :barnum 1 :swings (list (create-swing 0 2)
                                                                                     (create-swing 1 1))
                                                             :num-lines 2))
                                       5)
              (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                            (create-swing 1 2))
                                    :num-lines 2)
                    (make-bar-swing :barnum 1 :swings (list (create-swing 0 2)
                                                            (create-swing 1 1))
                                    :num-lines 2)))


#||

Create swing multiplication list
================================

Input:
=======
FilledoutSwing - list<pos, value>
Beats - list<pos, barnum, beatnum>


Output:
list<pos,tempo_multiplication,hold>

||#


(define-struct tempo-multiplier
  :y1 :x1
  :y2 :x2)


(define (create-tempo-multipliers-from-swing start-line bar-swing)
  (define swings (bar-swing :swings))
  (define num-lines (bar-swing :num-lines))
  (define end-line (+ start-line num-lines))
  (define weight-sum (let loop ((swings swings))
                       (if (null? swings)
                           0
                           (let* ((swing (car swings))
                                  (line1 (get-swing-line swing))
                                  (next-swing (cl-cadr swings))
                                  (line2 (if next-swing
                                             (get-swing-line next-swing)
                                             num-lines))
                                  (duration (- line2 line1)))
                             (+ (* duration 
                                   (/ 1 (get-swing-weight swing)))
                                (loop (cdr swings)))))))
  (let loop ((swings swings))
    (if (null? swings)
        '()
        (let* ((swing (car swings))
               (line1 (get-swing-line swing))
               (next-swing (cl-cadr swings))
               (line2 (if next-swing
                          (get-swing-line next-swing)
                          num-lines))
               (duration (- line2 line1)))
          (define multiplier (/ (* duration
                                   (/ 1 (get-swing-weight swing))
                                   num-lines)
                                weight-sum))
          (cons (make-tempo-multiplier :y1 (+ start-line line1) :x1 multiplier
                                       :y2 (+ start-line line2) :x2 multiplier)
                (loop (cdr swings)))))))

(***assert*** (create-tempo-multipliers-from-swing 10 (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                              (create-swing 1 2))
                                                                      :num-lines 2))
              (list (make-tempo-multiplier :y1 10 :x1 4/3
                                           :y2 11 :x2 4/3)
                    (make-tempo-multiplier :y1 11 :x1 2/3
                                           :y2 12 :x2 2/3)))
              
(***assert*** (create-tempo-multipliers-from-swing 10 (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                              (create-swing 1 2))
                                                                      :num-lines 3))
              (list (make-tempo-multiplier :y1 10 :x1 3/2
                                           :y2 11 :x2 3/2)
                    (make-tempo-multiplier :y1 11 :x1 6/4
                                           :y2 13 :x2 6/4)))
              

#!!
(pp (create-tempo-multipliers-from-swing 10 (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                    (create-swing 1 2))
                                                            :num-lines 3)))
(pp (create-tempo-multipliers-from-swing 10 (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                    (create-swing 1 2)
                                                                                    (create-swing 2 2))
                                                            :num-lines 3)))
!!#

;; TODO:
;; * Call create-filled-out-bar-swings
;; * Add (:x1 1 :x1 1) for bars that has no defined bar-swing
(define (create-tempo-multipliers-from-swings bars bar-swings num-lines)
  (define bar (cl-car bars))
  (define bar-swing (cl-car bar-swings))
  (define line (and bar (get-first-line-in-bar bar)))

  (if (not bar-swing)      
      (if bar
          (list (make-tempo-multiplier :y1 line :x1 1
                                       :y2 num-lines :x1 1))
          '())
      (append (create-tempo-multipliers-from-swing line bar-swing)
              (create-tempo-multipliers-from-swings (cdr bars)
                                                    (cdr bar-swings)
                                                    num-lines))))

#!!
(pp (create-tempo-multipliers-from-swings (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                                (make-bar :barnum 1 :beats (list (create-beat 2 0)))
                                                (make-bar :barnum 2 :beats (list (create-beat 4 0))))
                                          (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                                                        (create-swing 1 2))
                                                                :num-lines 2)
                                                (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                        (create-swing 1 2))
                                                                :num-lines 2)
                                                (make-bar-swing :barnum 2 :swings (list (create-swing 0 2)
                                                                                        (create-swing 1 1))
                                                                :num-lines 2))
                                          64))
!!#

#!!
(***assert*** (create-tempo-multipliers-from-swings (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                                          (make-bar :barnum 1 :beats (list (create-beat 2 0)))
                                                          (make-bar :barnum 2 :beats (list (create-beat 4 0))))
                                                    (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                                                                  (create-swing 1 2))
                                                                          :num-lines 2)
                                                          (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                                  (create-swing 1 2))
                                                                          :num-lines 2)
                                                          (make-bar-swing :barnum 2 :swings (list (create-swing 0 2)
                                                                                                  (create-swing 1 1))
                                                                          :num-lines 2)))
              (list (make-tempo-multiplier :y1 0 :multiplier 4/3 :multiplier-end 4/3)
                    (make-tempo-multiplier :y1 1 :multiplier 2/3 :multiplier-end 2/3)
                    (make-tempo-multiplier :y1 2 :multiplier 4/3 :multiplier-end 4/3)
                    (make-tempo-multiplier :y1 3 :multiplier 2/3 :multiplier-end 2/3)
                    (make-tempo-multiplier :y1 4 :multiplier 2/3 :multiplier-end 2/3)
                    (make-tempo-multiplier :y1 5 :multiplier 4/3 :multiplier-end 4/3)))
!!#



;;;;;;;;;;;;;;;;;;;;;;;

#||
(define (merge-tempo-multipliers mults1 mults2)

  (define (get-in-between-multiplier multiplier-value line prev-mult next-mult)
    (if (not prev-mult)
        multiplier-value
        (begin
          (assert (> line (prev-mult :line)))
          ;;(assert (< line (next-mult :line)))
          (cond ((= (prev-mult :multiplier)
                    (prev-mult :multiplier-end))
                 (* multiplier-value (prev-mult :multiplier)))
                ((not next-mult)
                 (throw "A non-hold tempo multiplier node can't be the last one"))
                (else
                 (* multiplier-value
                    (scale line 
                           (prev-mult :line) (next-mult :line)
                           (prev-mult :multiplier) (prev-mult :multiplier-end))))))))
  
  (let loop ((mults1 mults1)
             (mults2 mults2)

             (prev-mult1 #f)
             (prev-mult2 #f))
    
    (define mult1 (cl-car mults1))
    (define mult2 (cl-car mults2))
    
    (cond ((and (not mult1)
                (not mult2))
           '())
          
          ((not mult1)
           (loop mults2
                 mults1

                 prev-mult2
                 prev-mult1))
          
          ((not mult2)
           (c-display "mults1/mults2" mults1 "\n" mults2)
           (c-display "prev-mult2" prev-mult2)
           (if prev-mult2
               (begin
                 (define prev-multiplier (prev-mult2 :multiplier))
                 (if (not (= prev-multiplier (prev-mult2 :multiplier-end)))
                     (throw "A non-hold tempo multiplier node can't be the last one"))
                 (map (lambda (mult)
                        (copy-tempo-multiplier mult
                                               :multiplier (* (mult :multiplier)
                                                              prev-multiplier)
                                               :multiplier-end (* (mult :multiplier-end)
                                                                  prev-multiplier)))
                      mults1))
               mults1))
          
          (else
           (let ((line1 (mult1 :line))
                 (line2 (mult2 :line))
                 (multiplier1 (mult1 :multiplier))
                 (multiplier2 (mult2 :multiplier)))
             
             (cond ((< line2 line1)
                    (loop mults2
                          mults1
                          prev-mult2
                          prev-mult1))
                   
                   ((< line1 line2)
                    (cons (make-tempo-multiplier :line line1
                                                 :multiplier (get-in-between-multiplier multiplier1
                                                                                        line1
                                                                                        prev-mult2
                                                                                        mult2)
                                                 :multiplier-end (- (get-in-between-multiplier 1 ;;(mult1 :multiplier-end)
                                                                                               line2
                                                                                               (mult1 :multiplier-end)
                                                                                               (cl-cadr mults1))))
                          (loop (cdr mults1)
                                mults2
                                
                                mult1
                                prev-mult2)))
                   
                   (else
                    (cons (make-tempo-multiplier :line line1
                                                 :multiplier (* multiplier1 multiplier2)
                                                 :multiplier-end (get-in-between-multiplier (mult1 :multiplier-end)
                                                                                            line2
                                                                                            prev-mult2
                                                                                            mult2))
                          (loop (cdr mults1)
                                (cdr mults2)

                                mult1
                                mult2)))))))))


(pp (merge-tempo-multipliers (list (make-tempo-multiplier :line 0
                                                          :multiplier 3
                                                          :multiplier-end 4)
                                   (make-tempo-multiplier :line 4
                                                          :multiplier 4
                                                          :multiplier-end 4))
                             (list (make-tempo-multiplier :line 1
                                                          :multiplier 5
                                                          :multiplier-end 6)
                                   (make-tempo-multiplier :line 5
                                                          :multiplier 6
                                                          :multiplier-end 6))))

(begin
  (define new_a_x (* (scale y
                            a_y1 a_y2
                            a_x1 a_x2)
                     (scale y
                            b_y1 b_y2
                            b_x1 b_x2)))
  (define new_a_x (* (scale y
                            (a :y1) (a :y2)
                            (a :x1) (a :x2))
                     (scale y
                            b_y1 b_y2
                            b_x1 b_x2))))

1. Translate both graphs so that they cover all lines
2. Merge them

(define (create-tempo-multiplier x1 y1 x2 y2)
  (list x1 y1 x2 y2))

(define tempo-multiplier-x1 car)
(define tempo-multiplier-y1 cadr)
(define tempo-multiplier-x2 caddr)
(define tempo-multiplier-y2 cadddr)

||#

(define (make-tempo-multipliers-cover-all-lines as max-y)
  (if (null? as)
      (list (make-tempo-multiplier :y1 0 :x1 1
                                   :y2 max-y :x2 1))
      (let ((as (let ((y1 ((car as) :y1)))
                  (if (> y1 0)
                      (cons (make-tempo-multiplier :y1 0 :x1 1
                                                   :y2 y1 :x2 1)
                            as)
                      as))))
        (let* ((l (last as))
               (y2 (l :y2)))
          (assert (<= y2 max-y))
          (if (= y2 max-y)
              as
              (append as (list (make-tempo-multiplier :y1 y2    :x1 1
                                                      :y2 max-y :x2 1))))))))

(define (merge-tempo-multipliers as bs max-y)
  (let loop ((as (make-tempo-multipliers-cover-all-lines as max-y))
             (bs (make-tempo-multipliers-cover-all-lines bs max-y))
             (y1 0))
             
    (if (= y1 max-y)
        '()
        (let* ((a (car as))
               (b (car bs))
               (a_y2 (a :y2))
               (b_y2 (b :y2)))
          
          (cond ((<= a_y2 y1)
                 (loop (cdr as)
                       bs
                       y1))

                ((<= b_y2 y1)
                 (loop as
                       (cdr bs)
                       y1))

                (else
                 (define (get-x y)
                   (* (scale y
                             (a :y1) (a :y2)
                             (a :x1) (a :x2))
                      (scale y
                             (b :y1) (b :y2)
                             (b :x1) (b :x2))))
                 (define y2 (min a_y2 b_y2))
                 (cons (make-tempo-multiplier :y1 y1 :x1 (get-x y1)
                                              :y2 y2 :x2 (get-x y2))
                       (loop as bs y2))))))))

;; test 1: same start and stop
(***assert*** (merge-tempo-multipliers (list (make-tempo-multiplier :y1 0 :x1 3
                                                                    :y2 4 :x2 4))
                                       (list (make-tempo-multiplier :y1 0 :x1 5
                                                                    :y2 4 :x2 6))
                                       4)
              (list (make-tempo-multiplier :y1 0 :x1 (* 3 5)
                                           :y2 4 :x2 (* 4 6))))

;; test 2: same start and stop, but does not cover all lines
(***assert*** (merge-tempo-multipliers (list (make-tempo-multiplier :y1 0 :x1 3
                                                                    :y2 4 :x2 4))
                                       (list (make-tempo-multiplier :y1 0 :x1 5
                                                                    :y2 4 :x2 6))
                                       8)
              (list (make-tempo-multiplier :y1 0 :x1 (* 3 5)
                                           :y2 4 :x2 (* 4 6))
                    (make-tempo-multiplier :y1 4 :x1 1
                                           :y2 8 :x2 1)))

;; test 3: a starts later than b
(***assert*** (merge-tempo-multipliers (list (make-tempo-multiplier :y1 1 :x1 3
                                                                    :y2 4 :x2 4))
                                       (list (make-tempo-multiplier :y1 0 :x1 5
                                                                    :y2 4 :x2 6))
                                       8)
              (list (make-tempo-multiplier :y1 0 :x1 5
                                           :y2 1 :x2 (scale 1 0 4 5 6))
                    (make-tempo-multiplier :y1 1 :x1 (* 3 (scale 1 0 4 5 6))
                                           :y2 4 :x2 (* 4 6))
                    (make-tempo-multiplier :y1 4 :x1 1
                                           :y2 8 :x2 1)))

;; test 4: b stops later than a
(***assert*** (merge-tempo-multipliers (list (make-tempo-multiplier :y1 0 :x1 3
                                                                    :y2 4 :x2 4))
                                       (list (make-tempo-multiplier :y1 0 :x1 5
                                                                    :y2 5 :x2 6))
                                       8)
              (list (make-tempo-multiplier :y1 0 :x1 (* 3 5)
                                           :y2 4 :x2 (* 4 (scale 4 0 5 5 6)))
                    (make-tempo-multiplier :y1 4 :x1 (scale 4 0 5 5 6)
                                           :y2 5 :x2 6)
                    (make-tempo-multiplier :y1 5 :x1 1
                                           :y2 8 :x2 1)))

;; test 5: a starts and stops before b, but they overlap
(***assert*** (merge-tempo-multipliers (list (make-tempo-multiplier :y1 0 :x1 3
                                                                    :y2 4 :x2 4))
                                       (list (make-tempo-multiplier :y1 1 :x1 5
                                                                    :y2 5 :x2 6))
                                       8)
              (list (make-tempo-multiplier :y1 0 :x1 3
                                           :y2 1 :x2 (scale 1 0 4 3 4))
                    (make-tempo-multiplier :y1 1 :x1 (* 5 (scale 1 0 4 3 4))
                                           :y2 4 :x2 (* 4 (scale 4 1 5 5 6)))
                    (make-tempo-multiplier :y1 4 :x1 (scale 4 1 5 5 6)
                                           :y2 5 :x2 6)
                    (make-tempo-multiplier :y1 5 :x1 1
                                           :y2 8 :x2 1)))


;; test 6: a starts and stops before b, and they don't overlap
(***assert*** (merge-tempo-multipliers (list (make-tempo-multiplier :y1 0 :x1 3
                                                                    :y2 3 :x2 4))
                                       (list (make-tempo-multiplier :y1 4 :x1 5
                                                                    :y2 5 :x2 6))
                                       8)
              (list (make-tempo-multiplier :y1 0 :x1 3
                                           :y2 3 :x2 4)
                    (make-tempo-multiplier :y1 3 :x1 1
                                           :y2 4 :x2 1)                    
                    (make-tempo-multiplier :y1 4 :x1 (* 1 5)
                                           :y2 5 :x2 (* 1 6))
                    (make-tempo-multiplier :y1 5 :x1 1
                                           :y2 8 :x2 1)))


;; test 7: a starts before b, and b stops before a
(***assert*** (merge-tempo-multipliers (list (make-tempo-multiplier :y1 0 :x1 3
                                                                    :y2 5 :x2 4))
                                       (list (make-tempo-multiplier :y1 1 :x1 5
                                                                    :y2 4 :x2 6))
                                       8)
              (list (make-tempo-multiplier :y1 0 :x1 3
                                           :y2 1 :x2 (* 1 (scale 1 0 5 3 4)))
                    (make-tempo-multiplier :y1 1 :x1 (* 5 (scale 1 0 5 3 4))
                                           :y2 4 :x2 (* 6 (scale 4 0 5 3 4)))
                    (make-tempo-multiplier :y1 4 :x1 (* 1 (scale 4 0 5 3 4))
                                           :y2 5 :x2 (* 1 (scale 5 0 5 3 4)))
                    (make-tempo-multiplier :y1 5 :x1 1
                                           :y2 8 :x2 1)))




;;;;;;;;;;;;;;;;;;;;;;;

(define (get-duration num-lines samplerate lpb bpm tempo-multiplier)
  (/ (* num-lines 60 samplerate)
     (* lpb bpm tempo-multiplier)))

#||
(get-duration 64 48000 4 120 1)
(* 48000 8)
||#

(define (apply-tempo-multipliers-to-timings timings tempo-multipliers num-lines)
  timings)

#||
(***assert*** (apply-tempo-multipliers-to-timings (list 0 1 2)
                                                  (list (make-tempo-multiplier :y1 0 :x1 2
                                                                               :y2 2 :x2 4))
                                                  64)
              (list (list :y1 0 :x1 (scale 0 0 2 2 4)
                          :y2 1 :x2 (scale 1 0 2 2 4))
                    (list :y2 1 :x1 (scale 1 0 2 2 4)
                          :y2 2 :x2 (scale 2 0 2 2 4))
                    (scale 2 0 2 2 4))
              2)
||#

;; tempo automation lines needs to be split into two if start and end cross 1.0, before calling this function.
;; (not necessary with bpm or lpb (where would "1.0" be for those?))

(define (convert-tempo-multipliers-to-timings main-lpb main-bpm samplerate tempo-multipliers)
  (let loop ((tempo-multipliers tempo-multipliers)
             (time1 0))
    (if (null? tempo-multipliers)
        '()
        (let* ((a (car tempo-multipliers))
               (num-lines (- (a :y2) (a :y1)))
               (avg-tempo (/ (+ (a :x1) (a :x2)) 2))
               (time2 (+ time1 (get-duration num-lines samplerate main-lpb main-bpm avg-tempo))))
          (c-display "num-lines/avg-tempo/time2" num-lines avg-tempo time2)
          (cons (copy-tempo-multiplier a
                                       :x1 time1
                                       :x2 time2)
                (loop (cdr tempo-multipliers)
                      time2))))))


;;341333.3333

#||
(c-display (pp (convert-tempo-multipliers-to-timings 120 4 48000
                                                     (list (make-tempo-multiplier :y1 0 :x1 0.5
                                                                                  :y2 64 :x2 2))))
           "\n"
           (/ (* 8 48000) 2/2))
    

(c-display (pp (convert-tempo-multipliers-to-timings 120 4 48000
                                                     (list (make-tempo-multiplier :y1 0 :x1 0.5
                                                                                  :y2 32 :x2 1)
                                                           (make-tempo-multiplier :y1 32 :x1 1
                                                                                  :y2 64 :x2 2))))
           "\n"
           (/ (* 8 48000) 2/2))
||#  


(***assert*** (convert-tempo-multipliers-to-timings 120 4 48000
                                                    (list (make-tempo-multiplier :y1 0  :x1 1
                                                                                 :y2 64 :x2 2)))
              (list (make-tempo-multiplier :y1 0 :x1 0
                                           :y2 64 :x2 (/ (* 8 48000) (average 1 2)))))

(***assert*** (convert-tempo-multipliers-to-timings 120 4 48000
                                                    (list (make-tempo-multiplier :y1 0  :x1 2
                                                                                 :y2 64 :x2 1)))
              (list (make-tempo-multiplier :y1 0 :x1 0
                                           :y2 64 :x2 (/ (* 8 48000) (average 1 2)))))

(***assert*** (convert-tempo-multipliers-to-timings 120 4 48000
                                                    (list (make-tempo-multiplier :y1 0  :x1 1
                                                                                 :y2 32 :x2 3/2)
                                                          (make-tempo-multiplier :y1 32 :x1 3/2
                                                                                 :y2 64 :x2 2)))
              (let ((len-a (/ (* 4 48000) (average 1 3/2)))
                    (len-b (/ (* 4 48000) (average 3/2 2))))
                (list (make-tempo-multiplier :y1 0 :x1 0
                                             :y2 32 :x2 len-a)
                      (make-tempo-multiplier :y1 32 :x1 len-a
                                             :y2 64 :x2 (+ len-a len-b)))))
                                                         

(define (bpms-or-lpbs-to-tempo-multipliers main-bpm bpms num-lines value-key)
  (define (get-tempo-multiplier bpm)
    (/ bpm
       main-bpm))
  (let loop ((bpms bpms))
    (if (null? bpms)
        '()
        (let* ((bpm1 (car bpms))
               (bpm2 (cl-cadr bpms))
               (x1 (get-tempo-multiplier (bpm1 value-key)))
               (x2 (if (or (not bpm2)
                           (= (bpm1 :logtype) *logtype-hold*))
                       x1
                       (get-tempo-multiplier (bpm2 value-key)))))
          (cons (make-tempo-multiplier :y1 (bpm1 :place)
                                       :x1 x1
                                       :y2 (if bpm2
                                               (bpm2 :place)
                                               num-lines)
                                       :x2 x2)
                (loop (cdr bpms)))))))
                                               
(define (bpms-to-tempo-multipliers main-bpm bpms num-lines)
  (bpms-or-lpbs-to-tempo-multipliers main-bpm bpms num-lines :bpm))

(define (lpbs-to-tempo-multipliers main-lpb lpbs num-lines)
  (bpms-or-lpbs-to-tempo-multipliers main-lpb lpbs num-lines :lpb))


(***assert*** (bpms-to-tempo-multipliers 120
                                         (list (hash-table* :bpm 20
                                                            :place 1
                                                            :logtype *logtype-hold*))
                                         64)
              (list (make-tempo-multiplier :y1  1 :x1 (/ 20 120)
                                           :y2  64 :x2 (/ 20 120))))

(***assert*** (lpbs-to-tempo-multipliers 4
                                         (list (hash-table* :lpb 20
                                                            :place 1
                                                            :logtype *logtype-hold*))
                                         64)
              (list (make-tempo-multiplier :y1  1 :x1 (/ 20 4)
                                           :y2  64 :x2 (/ 20 4))))

(***assert*** (bpms-to-tempo-multipliers 120
                                         (list (hash-table* :bpm 20
                                                            :place 1
                                                            :logtype *logtype-hold*)
                                               (hash-table* :bpm 50
                                                            :place 8
                                                            :logtype *logtype-hold*))
                                         64)
              (list (make-tempo-multiplier :y1  1 :x1 (/ 20 120)
                                           :y2  8 :x2 (/ 20 120))
                    (make-tempo-multiplier :y1  8 :x1 (/ 50 120)
                                           :y2 64 :x2 (/ 50 120))))

(***assert*** (bpms-to-tempo-multipliers 120
                                         (list (hash-table* :bpm 200
                                                            :place 1
                                                            :logtype *logtype-linear*)
                                               (hash-table* :bpm 50
                                                            :place 8
                                                            :logtype *logtype-linear*)) ;; Logtype of last bpm is ignored.
                                         64)
              (list (make-tempo-multiplier :y1  1 :x1 (/ 200 120)
                                           :y2  8 :x2 (/ 50 120))
                    (make-tempo-multiplier :y1  8 :x1 (/ 50 120)
                                           :y2 64 :x2 (/ 50 120))))


;;(pp (<ra> :get-all-bpm))


(define (temponodes-to-tempo-multipliers temponodes)
  (let* ((node1 (car temponodes))
         (node2 (cadr temponodes))
         (x1 (node1 :tempo-multiplier))
         (x2 (if (= (node1 :logtype) *logtype-hold*)
                 x1
                 (node2 :tempo-multiplier))))
    (cons (make-tempo-multiplier :y1 (node1 :place)
                                 :x1 x1
                                 :y2 (node2 :place)
                                 :x2 x2)
          (if (null? (cddr temponodes))
              '()
              (temponodes-to-tempo-multipliers (cdr temponodes))))))

(***assert*** (temponodes-to-tempo-multipliers (list (hash-table* :place 0
                                                                  :tempo-multiplier 1
                                                                  :logtype *logtype-linear*)
                                                     (hash-table* :place 64
                                                                  :tempo-multiplier 1
                                                                  :logtype *logtype-linear*)))
              (list (make-tempo-multiplier :y1 0 :x1 1
                                           :y2 64 :x2 1)))

(***assert*** (temponodes-to-tempo-multipliers (list (hash-table* :place 0
                                                                  :tempo-multiplier 0.5
                                                                  :logtype *logtype-linear*)
                                                     (hash-table* :place 1
                                                                  :tempo-multiplier 0.1
                                                                  :logtype *logtype-linear*)
                                                     (hash-table* :place 64
                                                                  :tempo-multiplier 1.2
                                                                  :logtype *logtype-linear*)))
              (list (make-tempo-multiplier :y1 0 :x1 0.5
                                           :y2 1 :x2 0.1)
                    (make-tempo-multiplier :y1 1 :x1 0.1
                                           :y2 64 :x2 1.2)))

#||
(pp (<ra> :get-all-temponodes))
(<ra> :get-temponode-place 3)
||#




(define (create-block-timings blocknum)
  (define num-lines (<ra> :get-num-lines blocknum))
  '(define tempo-multipliers (merge-tempo-multipliers (merge-tempo-multipliers (bpms-to-tempo-multipliers (vector->list (<ra> :get-all-bpm blocknum)))
                                                                               (lpbs-to-tempo-multipliers (vector->list (<ra> :get-all-lpb blocknum)))
                                                                               num-lines)
                                                      (temponodes-to-tempo-multipliers (vector->list (<ra> :get-all-temponodes blocknum)))
                                                      
                                                      num-lines))
  (define tempo-multipliers '())
  (define samplerate (<ra> :get-sample-rate))
  (define plain-line-duration (get-duration 1
                                            samplerate 
                                            (<ra> :get-main-lpb)
                                            (<ra> :get-main-bpm)
                                            1))

  (define plain-timings (let loop ((line 0)
                                   (time 0))
                          (if (= line num-lines)
                              (list time)
                              (cons time
                                    (loop (1+ line)
                                          (+ time plain-line-duration))))))

  (apply-tempo-multipliers-to-timings plane-timings tempo-multipliers num-lines))



(when *is-initializing*
  (***assert*** (length (create-block-timings -1))
                64)
  (***assert*** (first (create-block-timings -1))
                0)
  (***assert*** (/ (last (create-block-timings -1))
                   (<ra> :get-sample-rate))
                8))



(define (place->stime-0 y
                        y1 y2
                        x1 x2)

  (define k (/ (- x2 x1)
               (- y2 y1)))

  (define d (- y2 y1))
  
  (define exponent (scale y y1 y2 0 k)) ;; number between 

  (define mult (expt d (* (- y y1)
                          (- y2 y1))))
                     
  
  (define time (+ y1 (* mult d)))

  (c-display "k" k "d" d "exp" exponent "mult" mult)
  
  time)


(list
 (place->stime-0 0.0001
                 0 10
                 0.1 1.0)
 (place->stime-0 5
                 0 10
                 0.1 1.0)
 (place->stime-0 10
                 0 10
                 0.1 1.0))

(expt 2 (* 5 1/12))


(***assert*** (place->stime-0 5
                              2 7
                              0.3 1.5)
              (let ((multiplier (scale 5 2 7 0.3 1.5)))
                (define avgmult1 (average 0.3 multiplier))
                (define avgmult2 (average multiplier 1.5))
                (define time1 (+ 


Her er problemet:

y = f(y0, y1, y2, x1, x2)

y0 = tid
y1 = start-tid
y2 = slutt-tid
x1 = tempo-multiplier ved start-tid
x2 = tempo-multiplier ved slutt-tid

Om x1==x2, altså at det er ingen tempo-forskjell, så skal funksjonen returnere y0.

Om x1==10, og x2==1, altså at tempoet gradvis minker, så skal funksjonen returnere et tall mellom y1 og y0.

Om x1==1, og x2==10, altså at tempoet gradvis øker, så skal funksjonen returnere et tall mellom y0 og y2.


Trengs det differensialligninger for å løse dette?


;; Linear tempo change.
(define (place->stime-0 y
                        y1 y2
                        x1 x2
                        gakk
                        )
  (define x (scale y y1 y2 x1 x2))
  (c-display "x" x "logy" (log (/ x x1)) "logy2" (log (/ x2 x1)))
  (scale (log (/ x x1))
         0 (log (/ x2 x1))
         y1 y2))

;  duration)

(list
 (place->stime-0 0
                 0 10
                 1.0 0.1)
 (place->stime-0 2.5
                 0 10
                 1.0 0.1)
 (place->stime-0 2.5
                 0 10
                 0.1 1.0)
 (place->stime-0 5
                 0 10
                 1.0 0.1)
 (place->stime-0 5
                 0 10
                 0.1 1.0)
 (place->stime-0 7.5
                 0 10
                 1.0 0.1)
 (place->stime-0 7.5
                 0 10
                 0.1 1.0)
 (place->stime-0 10
                 0 10
                 0.1 1.0))

(define (equal-ratios bp          ;; beat
                      T1 T2       ;; start tempo, end tempo
                      B)          ;; total beats

  (define Tbp (* T1 (expt (/ T2 T1) (/ bp B)))) ;; Tempo at bp
  (define DURbp (* (/ 60 T1)
                   (/ B (log (/ T1 T2)))
                   (- (expt (/ T1 T2) (/ bp B))
                      1)))

  (c-display "Tbp" Tbp)
  DURbp)

(list (equal-ratios 0 -80 -160 16)
      (equal-ratios 4 80 160 16)
      (equal-ratios 8 80 160 16)
      (equal-ratios 12 80 160 16)
      (equal-ratios 16 80 160 16))

;; exponential tempo change ("equal ratios")      
(define (place->stime-0 y
                        y1 y2
                        x1 x2
                        gakk)
  (define bp y)         ;; beat
  (define T1 x1)        ;; start tempo
  (define T2 x2)        ;;end tempo
  (define B  (- y2 y1))          ;; total beats

  (if gakk ;;(< x2 x1))
      (begin
        
        (define DLT
          (begin
            (define x (scale y y1 y2 x1 x2))
            ;;(c-display "x" x "logy" (log (/ x x1)) "logy2" (log (/ x2 x1)))
            (scale (log (/ x x1))
                   0 (log (/ x2 x1))
                   y1 y2)))
        
        (define Tbp (+ T1 (- T2
                             (* T1 (expt (/ T2 T1) (/ (- B bp) B)))))) ;; Tempo at bp
        (define DURbp (* (/ 1 T1) ;;(/ 60 T1)
                         (/ B (log (/ T1 T2)))
                         (- (expt (/ T1 T2) (/ bp B))
                            1)))
        ;;(define DLT (- y y1)) ;; duration associated with a linear tempo change
        ;;(set! DLT (- y y1))
        (define DER DURbp)
        
        (c-display "Tbp" Tbp)
        (- DLT (- DER DLT))
        
        ;;        Tbp
        ;;DLT
        ;;(scale y y1 y2 x1 x2)
        Tbp
        DLT
        )
      
      (begin

        (define Tbp (* T1 (expt (/ T2 T1) (/ bp B)))) ;; Tempo at bp
        (define DURbp (* (/ 1 T1) ;;(/ 60 T1)
                         (/ B (log (/ T1 T2)))
                         (- (expt (/ T1 T2) (/ bp B))
                            1)))
        
        (c-display "Tbp" Tbp)
        DURbp
        ;;Tbp
        )))


;; exponential tempo change ("equal ratios"), only ritardando
(define (place->stime-0 y
                        y1 y2
                        x1 x2
                        gakk)
  (define a y1)
  (define b y2)
  (define r y);;(- b a))
  (define Ta x1)        ;; start tempo
  (define Tb x2)        ;;end tempo
  
  (define xr0 a)
  (define xr (/ (- r a)
                (- b a)))
  
  (define e 2.71828182845904523536028747135266249775724709369995)

  (define first (/ (- b a)
                   (* Tb (log (/ Tb Ta)))))
  (define second_a (expt e (* (log (/ Tb Ta))
                              (- 1 xr))))
  (define second_b (expt e (* (log (/ Tb Ta))
                              (- 1 xr0))))

  (define second (- second_a second_b))

  (* first second)

  ;; return tempo instead
  (c-display "xr" xr "Ta" Ta "Tb" Tb (> Tb Ta))
  
  (define use-icmc-formula #f)

  (if use-icmc-formula
      (if (> Tb Ta)
          (* Ta
             (expt (/ Tb Ta)
                   xr))
          (+ Ta
             (- Tb (* Ta
                      (expt (/ Tb Ta)
                            (/ (- b r)
                               (- b a)))))))
      (if #f
          (* Ta
             (expt (/ Tb Ta)
                   xr))
          (* Tb
             (expt (/ Ta Tb)
                   (- 1 xr)))))
  

  ;;(* 0.1 (expt 52 (1- xr)))
  )

t1 + ( t2 - (t1 * (t2/t1)^( (10-x) / 10)))

integrate (t1 + ( t2 - (t1 * (t2/t1)^( (c-x) / c))))^-1, x from 0 to b

y(x) = 5.2 + ( 0.1 - (5.2 * (0.1/5.2)^( (10-x) / 10)))


(define (draw-plot xs func)
  (define ys (map func xs))
  (define args "")
  (for-each (lambda (x y)
              (set! args (<-> args " " (* 1.0 x) " " (* 1.0 y) )))
            xs ys)
  (system (<-> "plot" args "&"))
  ;;(read-char)
  (c-display "args" args)
  ys)

 
(define g1 (draw-plot (map (lambda (x) (/ (1+ x) 10)) (iota 100))
                      (lambda (x)
                        (place->stime-0 x
                                        0 10
                                        5.2 0.1
                                        #t)))
  )

(define g2 (draw-plot (map (lambda (x) (/ (1+ x) 10)) (iota 100))
                      (lambda (x)
                        (place->stime-0 x
                                        0 10
                                        0.1 5.2
                                        #f)))
  )


(list
 (place->stime-0 0
                 0 10
                 0.5 1.25)
 (place->stime-0 2.5
                 0 10
                 0.5 1.25)
 (place->stime-0 5
                 0 10
                 0.5 1.25)
 (place->stime-0 7.5
                 0 10
                 0.5 1.25)
 (place->stime-0 10
                 0 10
                 0.5 1.25))
      


(expt 2 10)

#||

Input:
=======

DefaultLPB - integer
DefaultBPM - integer
LPB - list<pos,value,is_hold>
BPM - list<pos,value,is_hold>
TempoAutomation - list<pos,tempoMul>, with at least two nodes. The first node has position 0, the last node has position NumLines.
Swing - list<?>
<strike>NumLines - integer</strike>
<strike>Samplerate - integer</strike>


Output:
=======

<strike>Timing - list, with at least one node per integer. First node has the value 0. Length is NumLines + 1.</strike>
Tempo automation: list<pos,tempo multiplication, is_hold>


Swing:
=======

0 -> 1
1 -> 2
2 -> 1
3 -> 2
4 -> 1
5 -> 2
6 -> 1
 ....
 ===>
0 -> 1.33
1 -> 0.66
2 -> 1.33
3 -> 0.66
 ....


LPB:
======
0 -> 1
1 -> 2
2 -> 1
3 -> 2
4 -> 1
5 -> 2
6 -> 1
 ....
 ===>
0 -> 1   (1/1
1 -> 0.5 (1/2)
2 -> 1   (1/2)
3 -> 0.5 (1/1
 ....


BPM:
======
0 -> 100
1 -> 200
2 -> 100
3 -> 200
4 -> 100
5 -> 200
6 -> 100
 ....
 ===>
0 -> 60/100
1 -> 60/200
2 -> 60/100
3 -> 60/200
 ....



(define (get-duration duration-in-lines samplerate lpb bpm tempo-multiplier)
  (/ (* duration-in-lines 60 samplerate)
     (* lpb bpm tempo-multiplier)))
  

(***assert*** (/ (get-duration 64 48000 4 120 1)
                 48000)
              8)

(define (create-posval pos val)
  (cons pos val))

(define (posval-pos posval)
  (car posval))

(define (posval-val posval)
  (cdr posval))

;; Could we add swing in the same type of format as lpb and bpm?
;; In case, maybe we could merge lpb/bpm/swing before calling this function.
(define (create-timing num-lines samplerate default-lpb lpbs default-bpm bpms)
  (let loop ((pos 0)
             (lpb default-lpb)
             (lpbs lpbs)
             (bpm default-bpm)
             (bpms bpms))
    (define lpb1 (cl-car lpbs))
    (define bpm1 (cl-car bpm1))
    (define next-line (1+ (floor pos)))

    (cond ((> pos num-lines)
           (assert (null? lpbs))
           (assert (null? bpms))
           '())

          ;; something like this.
          ((and lpb (<= (posval-pos lpb1)
                        next-line))
           (define lpb1pos (posval-pos lpb1))
           (cons (+ last-time
                    (get-duration (- lpb1pos last-pos)
                                  samplerate
                                  lpb
                                  pbm
                                  1.0))
                 (loop lpb1pos
                       (posval-value lpb1)
                       (cdr lpbs)
                       bpm
                       bpms)))

          (else
           (cons (get-duration line samplerate default-lpb default-bpm 1.0)
                 (loop next-line
                       lpbs
                       bpms))))))


(let ((num-lines 2)
      (samplerate 48000)
      (default-lpb 4)
      (lpbs '())
      (default-bpm 125)
      (bpms '()))
  (***assert*** (create-timing num-lines samplerate default-lpb lpbs default-bpm bpms)
                (list 0
                      (get-duration 1 samplerate default-lpb default-bpm 1)
                      (get-duration 2 samplerate default-lpb default-bpm 1))))

(/ (get-time 64 48000 4 120 .5)
   48000.0)

||#


